/*
 * This file is part of MinimalGS, which is a GameScript for OpenTTD
 * Copyright (C) 2011  Leif Linse
 *
 * MinimalGS is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * MinimalGS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MinimalGS; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

/* Import SuperLib for GameScript */
import("util.superlib", "SuperLib", 24);
Log <- SuperLib.Log;
Helper <- SuperLib.Helper;

/* Import TileLabels */
import("scenario.tilelabels", "TileLabels", 1);

require("version.nut");
require("version_upgrade.nut");

/* Globals */
g_tile_labels <- null;
g_scp <- null;

class MainClass extends GSController 
{
	_load_data = null;

	_raise1_tile = null;
	_raise2_tile = null;

	_condition_industries = null;
	_condition_min_total_production = null;
	_condition_min_period = null;

	// record the date when the condition first was met
	// if the condition is no longer met before a period of X months,
	// this variable will be reset to null.
	_condition_met_from_date = null;

	_goal_accomplished = null;
	_goal_id = null;

	constructor()
	{
		this._load_data = null;

		this._raise1_tile = null;
		this._raise2_tile = null;

		this._condition_industries = [];
		this._condition_min_total_production = null;
		this._condition_met_from_date = null;
		this._condition_min_period = null;
		this._goal_accomplished = false;
		this._goal_id = null;
	}
}

function MainClass::Start()
{
	this.Init();
	this.Welcome(0);

	// Wait for the game to start
	this.Sleep(1);

	local last_check = 0;

	while (true) {
		local loop_start_tick = GSController.GetTick();

		this.HandleEvents();

		local now = GSDate.GetCurrentDate();
		if(last_check + 15 < now &&
				GSDate.GetDayOfMonth(now) < 5)
		{
			last_check = now;
			this.Monthly();
		}

		this.Daily();

		// Loop with a frequency of one day
		local ticks_used = GSController.GetTick() - loop_start_tick;
		this.Sleep(Helper.Max(1, 5 * 74 - ticks_used));
	}
}

function MainClass::Init()
{
	// Initialize the tile labels library
	g_tile_labels = TileLabels("$L=");
	if (this._load_data != null && this._load_data.rawin("tile_labels"))
	{
		g_tile_labels.ImportFromSave(this._load_data.tile_labels);
	}
	g_tile_labels.ReadFromMap();

	if(this._load_data != null)
	{
		if(this._load_data.rawin("goal_id"))
			this._goal_id = this._load_data.goal_id;

		if(this._load_data.rawin("condition_met"))
			this._goal_accomplished = this._load_data.condition_met;

		if(this._load_data.rawin("condition_met_from_date"))
			this._condition_met_from_date = this._load_data.condition_met_from_date;
	}

	this.SetIdleMoreCompanyName();

	// Read data from TileLabels
	this._raise1_tile = g_tile_labels.GetTile("raise1");
	this._raise2_tile = g_tile_labels.GetTile("raise2");

	local i = 1;
	while(i < 100)
	{
		local tile = g_tile_labels.GetTile("industry" + i);
		if(!GSMap.IsValidTile(tile))
			break;

		local industries = GSIndustryList();
		industries.Valuate(GSIndustry.GetDistanceManhattanToTile, tile);
		industries.KeepBottom(1);
		local industry = industries.Begin();

		this._condition_industries.append(industry);
		i++;
	}

	this.ReadGSSettings();

	// Only create the goal if it has not yet been created (and goal has not been met)
	if(this._goal_id == null && (this._goal_accomplished == null || this._goal_accomplished == false))
		this._goal_id = GSGoal.New(GSCompany.COMPANY_INVALID, GSText(GSText.STR_GOAL, this._condition_min_total_production, this._condition_min_period), GSGoal.GT_NONE, 0);

}

function MainClass::ReadGSSettings()
{
	this._condition_min_total_production = GSController.GetSetting("min_total_production");
	this._condition_min_period = GSController.GetSetting("min_period");
}

function MainClass::SetIdleMoreCompanyName()
{
	// Rename the AI company so its more obvious to veteran 
	// players/developers that they shouldn't remove it.
	local cm = GSCompanyMode(1);
	GSCompany.SetName("Deity slave company");
	GSCompany.SetPresidentName("Do not stop this AI");
}

function MainClass::Save()
{
	Log.Info("Saving data to savegame", Log.LVL_DEBUG);
	return { 
		tile_labels = g_tile_labels.ExportToSave(),
		goal_id = this._goal_id,

		// for historic reasons this variable is stored as "condition_met" 
		// in the save
		condition_met = this._goal_accomplished 

		condition_met_from_date = this._condition_met_from_date
	};
}

function MainClass::IntToStrFillAfter(intVal, desiredStrLen)
{
	local str = "" + intVal;

	while(str.len() < desiredStrLen)
	{
		str += " ";
	}
	
	return str;
}

function MainClass::Load(version, tbl)
{
	Log.Info("Loading data from savegame of game script version: " + version, Log.LVL_INFO);

	if(!ALLOW_UPGRADE && version != SELF_VERSION)
	{
		GSLog.Error("------------------------------------------------------------");
		GSLog.Info("                                                           ");
		GSLog.Info("    Error: You do not have the correct version of the      ");
		GSLog.Info("           game script that is required to play this       ");
		GSLog.Info("           scenario.                                       ");
		GSLog.Info("                                                           ");
		GSLog.Info("           You have version " + SELF_VERSION + ",          ");
		GSLog.Info("           but need version " + version + "                ");
		GSLog.Info("                                                           ");
		GSLog.Info("    If you just downloaded the scenario from online        ");
		GSLog.Info("    content download, please go back and ensure that       ");
		GSLog.Info("    you also download the last version of                  ");
		GSLog.Info("    SplitScenarioGS which is required for the scenario     ");
		GSLog.Info("    to work.                                               ");
		GSLog.Info("                                                           ");
		GSLog.Info("    If you still can't resolve this issue there might      ");
		GSLog.Info("    have been an error in the packaging of the             ");
		GSLog.Info("    scenario. Pleases report errors that you believe       ");
		GSLog.Info("    is caused by this to the forum thread of the           ");
		GSLog.Info("    scenario:                                              ");
		GSLog.Info("                                                           ");
		GSLog.Info("    http://www.tt-forums.net/viewtopic.php?f=60&t=61378    ");
		GSLog.Info("                                                           ");
		GSLog.Error("------------------------------------------------------------");
		GSLog.Info("");

		throw "Please read the error message above.";
	}

	// Store a copy of the table from the save game
	this._load_data = {}
   	foreach(key, val in tbl)
	{
		this._load_data.rawset(key, val);
	}	
}

function MainClass::HandleEvents()
{
	if(GSEventController.IsEventWaiting())
	{
		local ev = GSEventController.GetNextEvent();

		if(ev == null)
			return;

		local ev_type = ev.GetEventType();
		if(ev_type == GSEvent.ET_COMPANY_NEW)
		{
			local company_ev = GSEventCompanyNew.Convert(ev);
			local company_id = company_ev.GetCompanyID();

			this.Welcome(company_id);
		}
	}
}

/* This function is called daily by the main loop */
function MainClass::Daily()
{
}

/* This function is called at the beginning of every month by the main loop */
function MainClass::Monthly()
{
	if(this._goal_accomplished)
		return; // Condition already fulfilled => no need to check it anymore

	// makes debugging easier if the goals can be changed during
	// game without save + load
	this.ReadGSSettings(); 

	local total_production = 0;
	local cargo_list = GSCargoList();
	foreach(industry in this._condition_industries)
	{
		foreach(cargo, _ in cargo_list)
		{
			local production = GSIndustry.GetLastMonthProduction(industry, cargo);

			if(production < 0) // don't report negative production towards the sum
				production = 0;

			total_production += production;

			Log.Info("production of " + GSCargo.GetCargoLabel(cargo) + " is: " + production + " ind: " + industry + " " + GSIndustry.GetName(industry), Log.LVL_DEBUG);
		}
	}

	Log.Info("Total production is: " + total_production + " (target: " + this._condition_min_total_production + ")", Log.LVL_INFO);

	// If production drops below the target, clear the start time
	// of the period during which the condition have been met
	if(this._condition_met_from_date != null &&
			total_production < this._condition_min_total_production)
	{
		this._condition_met_from_date = null;
		this.GoalPeriodAbortedMessage();
		Log.Info("Cleared start date of period during which the goal is met", Log.LVL_INFO); 
	}

	// Have we just reached a production high enough to start counting
	// a period with high enough production to satisfy the goal
	if(total_production > this._condition_min_total_production &&
		this._condition_met_from_date == null)
	{
		Log.Info("Recorded the start of a period during which the goal is met", Log.LVL_INFO); 
		this._condition_met_from_date = GSDate.GetCurrentDate();
		this.GoalPeriodStartedMessage(); // inform players.
	}

	// Have the condition been met for a period of at least 
	// _condition_min_period months? (ignore that not all months are 30
	// days. Unless this._condition_min_period is very large, it will not
	// be noticed)
	if(this._condition_met_from_date != null &&
		this._condition_met_from_date + this._condition_min_period * 30 < GSDate.GetCurrentDate())
	{
		Log.Info("Goal have been accomplished", Log.LVL_INFO); 
		// Yes, the goal have been accomplished.

		this._goal_accomplished = true;
		if(this._goal_id != null)
		{
			GSGoal.Remove(this._goal_id);
			this._goal_id = null;
		}
		this.GoalCompleteMessage(); // inform players.
		this.RaiseLand();
	}
}

// Called when the condition has been met for raising land
function MainClass::RaiseLand()
{
	Log.Info("raise land", Log.LVL_INFO);
	local cm = GSCompanyMode(1); // Switch to first AI, an IdleMore AI with lot of cheated money

	local tile_area = GSTileList();
	tile_area.AddRectangle(this._raise1_tile, this._raise2_tile);

	Log.Info("tile area size: " + tile_area.Count(), Log.LVL_INFO);

	// randomize the order that we process the tiles to give an interesting visual effect
	tile_area.Valuate(GSBase.RandItem);
	tile_area.Sort(GSList.SORT_BY_VALUE, GSList.SORT_ASCENDING);

	local corners = [GSTile.CORNER_N, GSTile.CORNER_E, GSTile.CORNER_S, GSTile.CORNER_W];
	local slopes = [GSTile.SLOPE_N, GSTile.SLOPE_E, GSTile.SLOPE_S, GSTile.SLOPE_W];
	foreach(tile, _ in tile_area)
	{
		local slopes_to_raise = 0;
		foreach(key, corner in corners)
		{
			if(GSTile.GetCornerHeight(tile, corner) == 0)
				slopes_to_raise = slopes_to_raise | slopes[key];
		}

		GSTile.RaiseTile(tile, slopes_to_raise);
		GSController.Sleep(4);
	}
}

// This function displays a welcome message to the given company
function MainClass::Welcome(company_id)
{
	if(GSController.GetSetting("display_welcome_message") == 1)
	{
		GSGoal.Question(0, 
				company_id, 
				GSText(GSText.STR_WELCOME, this._condition_min_total_production, this._condition_min_period, this._condition_min_period), 
				GSGoal.QT_INFORMATION,
				GSGoal.BUTTON_CLOSE);
	}
}

function MainClass::DisplayNewsForAllCompanies(message)
{
	for(local company_id = GSCompany.COMPANY_FIRST; company_id <= GSCompany.COMPANY_LAST; company_id++)
	{
		if(GSCompany.ResolveCompanyID(company_id) != GSCompany.COMPANY_INVALID)
		{
			GSNews.Create(GSNews.NT_GENERAL,
					message,
					company_id);
		}
	}
}

function MainClass::DisplayMessageForAllCompanies(message)
{
	for(local company_id = GSCompany.COMPANY_FIRST; company_id <= GSCompany.COMPANY_LAST; company_id++)
	{
		if(GSCompany.ResolveCompanyID(company_id) != GSCompany.COMPANY_INVALID)
		{
			GSGoal.Question(0, 
					company_id, 
					message, 
					GSGoal.QT_INFORMATION,
					GSGoal.BUTTON_CLOSE);
		}
	}
}

function MainClass::GoalPeriodStartedMessage()
{
	MainClass.DisplayNewsForAllCompanies(
		GSText(GSText.STR_GOAL_PERIOD_STARTED, 
			this._condition_min_total_production, 
			this._condition_min_period));
}
function MainClass::GoalPeriodAbortedMessage()
{
	MainClass.DisplayNewsForAllCompanies(
		GSText(GSText.STR_GOAL_PERIOD_ABORTED, 
			this._condition_min_total_production));
}
function MainClass::GoalCompleteMessage()
{
	MainClass.DisplayMessageForAllCompanies(GSText(GSText.STR_COMPLETE));

	// In non-multiplayer games, scroll view to the land that will be raised
	if(!GSGame.IsMultiplayer())
	{
		GSViewport.ScrollTo(this._raise1_tile);
	}
}
